# CSSMinMaxR24.py for Notepad++ Python Scripting plugin * For CSS developers and web designers
# This is improved Python port of the YUI CSS Compressor featured to minify/format modern CSS
#
## Extends 2024 (R24)
# Patched by Barmaley: bugs fixed + useful features added * GPLv3 * GoodCRX Project
#
# 2024.8.31 (initial fork):
# 1. No required space for @media ") or (". Fixed.
# 2. No required spaces in calc with "+" sign 'calc(1px + 1em)'. Fixed.
# 3. Preserved comments now defined with /**..*/ (instead of original /*!..*/).
# 4. If some text selected in active file: process fragment minification in current document (as in original script, see Usage below).
# 5. If nothing selected in active file: force process minification in NEW document (create new than process the copy of source CSS).
# 6. After processing autosave* new compressed file, autonamed from original name with change to "_.css".
#	 * Note: if any file with same name exist in the same dir - it's overwrite without confirmation.
# 7. Force pre-save current focument backup copy* (create in same dir as opened file, add .BAK to current filename).
#	 * This copy overwrite every time. No way in NPP to automatically remove temp backup files. Delete them manually if need.
#
# 2024.8.30 -> 2024.8.31:
# + Compressed CSS: all preserved comments delimits by new line from previous/following CSS rules (\n).
# + Fix logic: reset UNDO ops ONLY if minified "by place" (if process run in current file, else undo-buffer stay invalid).
# + Fix/improved: regexp for "calc(x + y)" expressions and for and/or conditions (e.g. @media/@import etc.).
# + Improve/optimize: multi-to-one zeros conversion, multi-semicolon remover, zero-value with units (px/vh/..) fixer.
#
# 2024.8.31 -> 2024.9.4:
# + Rebrand project: nows it's named as CSSMinMaxR24.
# + Simplest CSS autoformatter: autodetect compressed CSS then expand it - every rules converted to indented CSS blocks:
# 	* If some text selected: replace formatted fragment "by place" (as compressed in previous versions).
# 	* If nothing selected: current compessed CSS expands, create new autonamed file with *_MAX.css as name.
#	* Yet works fine only with usual and one-level conditional rules (e.g. @media), otherwise text indents may be not "beauty"
# + Fix/improved: regexp for "calc(x + y)" expressions and for and/or conditions (e.g. @media/@import etc.).
# + Fix/improved: check/autofix for missing/excess :: in pseudo-elements (::after/etc) + fix excess :: in pseudo-classes (:hover/etc).
# + Improve/optimize: multi-to-one zeros conversion, multi-semicolon remover, zero-value with units (px/vh/..) fixer.
# + Simplest "CSS-beautifier": add some spaces after/before delimiters (e.g. ,_ :_ _{ ), no trailing spaces and empty lines.
# + Source code optimization: some "one-time-call" function moved to aproriate execution points.
#
# 2024.9.4 -> 2024.9.7:
# + Comment empty rules in compressor: rules with no declarations safe "self-backuped", no remove em (it will not processed by browser).
# + Fix/improved: check/autofix for missing/excess :: in pseudo-elements and excess ::+ in pseudo-classes (:hover/etc).
# + Post-process usability: after compress/expand - scroll to top if new file created or re-select converted fragment in current file.
# + Optional: line wrap in the compressor: get output CSS structure u want - "brick"/rules-by-lines/etc (is in but unused in source cssmin).
# + Options block added: u can check/change the "set{...}" params to adjust actions (read comments to understand param values).
#	Most of all options adjust the CSS-compressor (to minimize output file size), CSS-expander not required adjustment in general.
#	With changes the option flags you can:
#	* on/off initial backup (force save copy of current document);
#	* control method of empty rules operations* (keep/comment/remove);
#	* set your own "important" comments* mark (if u want preserv some comments), skip comments in actions (as-is) or force remove em;
#		* note: empty rules and comments can be removed ONLY by compressor, expander (CSS-formatter) NOT remove any trash anyway;
#	* control apply experimental features (e.g. pseudo-elements/classes correction);
#	* on/off wrap lines mode for compressor and/or beautifier for expander post-processing (improve readability);
#	* set your own mask of compressed/expanded CSS-file (when process entire document and autosave results in a new file);
#	* set every time autosaving before processing, intend only for converts "by place" (if works with selected text).
#
# 2024.9.7 -> 2024.9.15:
# + Fix/foolproof: convert direction autodetector fail correct detection in some case (for unlogic option flags combination). Fixed.
# + Fix/improved: check/autofix for missing :: in pseudo-elements (now check for all current MDN pe-items).
# + Fix/improved: processing of and/or in at-rules and math signs in calc() - check/fix spaces before/after that (both compressor/expander).
# + Force remove empty comment like /* */, anyway (both compress/expand ops, not intends of comments flag u set).
# + Force* move a @charset or @import rules to the BOF (both compress/expand ops). All @charset rules except the first will be force removed!
#	* See at //developer.mozilla.org/docs/Web/CSS/@charset and //developer.mozilla.org/docs/Web/CSS/@import for details.
# + Improved: conditional rules (CR) formatter now works fine with most* of multi-level CR (indents for @media/@supports/etc. inner rules).
#	* Note: sometimes u can get "crooked fir". It's possible in multi-level 4+ CR. Not bring bugs in output CSS, but sign to check such rules.
# + Optional: on/off multi-level* format of expanded CSS (expander only). If on - set extra indents in at-rules (@media/@supports/etc).
#	* If off - only basic CSS-format, indents will set only for in-rules declares, any embedded conditional rules processed as usual rules.
#	* As result - u get simplest one-level output CSS tree. Most fast processing with no iteratable steps to indent the inner conditions.
# + Optional: comments indent - compressor only (expander always do that), if on - all comments in output CSS will be indented by new line.
# + Optional: first comment protection, keep first* comment (CSS "cap") both in compressor/expander, stay unchanged and not removed anyway.
#	* Placed at BOF and NOT empty! Otherwise comment NOT consider as "first". Any preceding empty lines, tabs, whitespaces be stripped.
# + Optional: extra-fine compressor feats (multi-zeros to one, RGB-colors to HEX, shorten HEX colors, etc.) also can be used* in expander.
#	* Get the less size of expanded CSS and correct some typical exceeds (e.g. multi spaces, ";", etc.). If use - u get fine CSS at one pass.
#	* Experimental option (set[psel]) now works in pair with this. Check/fix pseudo-elements/classes now execute ONLY if "extra-fine" is ON.
# + The Box Model Hack processing in compressor now is optional* step (this can be used ONLY for old MSIE-5/6, useless for modern browsers).
#	* Default is off (set[bmhk]:0), two exseed cycles in compressor excluded, faster processing. Set on ONLY if your CSS has fixes for IE5/6.
# 	* See more about pseudo-class fix here: //tantek.com/CSS/Examples/boxmodelhack.html (RU: //www.webmascon.com/topics/coding/38a.asp).
# + Apropriate options flags added for new optional features: u can on/off em in the "set{...}" params block.
# + Misc: add exseed [.#>] corrector (rule definitions), add test CSS-file in package.
#
# 2024.9.15 -> 2024.9.30:
# + Improved: process direction autodetector - now CSS is "yet uncompressed" if at least one line-break after "{" symbol is in source.
# + Fix/improved: extended at-rules formatter - add "crutches" to align tab-shifted "crooked firs" in some problem rules inside CR.
# + Fix/improved: if CSS contains calc() declare - some cases in declare followed that excees whitespaces around "-" occured. Fixed.
# + Fix/improved: if extra-fine procs used - some pseudo-elements broken by wrong regexp, incorrect in-rule declares occured. Fixed.
# + Fix/improved: if prevent first comments used - any in-rules comments can become "first" and that ripped from parent rule. Fixed.
# + Fix/improved: if comments autoprotection used in compressor - some cases wrong conversion occured ("eat" some CSS parts). Fixed.
# + Fix/improved: if comments autoprotection used in compressor AND if latest comment was NOT closed - infinite circle occur. Fixed.
# + Optional: flag to process comments hacked* for old IE5 on Mac (set[ciem]). Now by default is off (not used in modern CSS).
#	* The hack to skip two comments around rule/group in IE5 on Mac: //www.oreilly.com/library/view/css-cookbook/0596005768/ch09s06.html
# + Optional: flag to save output CSS file as UTF-8 (set[utf8]). Default is on, get correct supports of non-latin symbols in CSS.
# + Improved: color condenser now support both 6/8-symbols hex-colors. Incorrect color pairs (e.g. #aabbccd) downshifted to hex-6.
# + Improved: color condenser now support any form of rgb(a)/hsl(a) colors. Converts to hex-6/8 (acc. 0.1%) with syntax errors correction.
# + Improved: color condenser now remove excess opacity bytes in hex-4/8 (if any abcF/abcdefFF exist) - the colors YET full with opacity 1.
# + Optional: convert "transparent" color to #0000 (set[tran]), vise versa in compressor/expander. Default is on, -6 bytes in compressed CSS.
# + Improved: "cosmetics" in line-wrapper (compressor only) - increase readability, any comments indents by line-breaks (except in-rules).
#
# 2024.9.30 -> 2024.10.4:
# + Improved: force protect literals (e.g. url('...'), content:"...", a:has("...")) - no punctuation shifts in compressor/formatter.
#
# 2024.10.4 -> 2024.10.5:
# + Fix/improved: template to detect full paths with ";" in "cap" block (possible literals in @import url(':;/{:;;..}'). Fixed.
# + Fix/improved: template to right detect first comment ("cap") - in some cases commented empty rules became the "cap". Fixed.
#
# 2024.10.5 -> 2024.10.31:
# + Fix/improved: beautifier regexp in expander - wrong whitespace placed between ")" and ":" in some rule definition occur. Fixed.
# + Fix/improved: extra finer zero compressor - some cases incorrect remove symbols after "0" (e.g. @keyframes...{0%{...}}). Fixed.
# + Fix/improved: protect spaces before ":" at rules definition in compressor - some cases that skipped and removes occured. Fixed.
# + Fix/improved: protect/add space before "(" in at-rules conditions after keyword "not" (if any) - was skipped or removed. Fixed.
# + Improved: process direction autodetector - now CSS also is "yet uncompressed" if at least one whitespace after "{" symbol is in source.
# + Improved: source UTF8 encode/decode now used ONLY if proceesed entire CSS (when output placed in new file, not "by place" conversion).
# + Improved: color condenser - special fix to convert quart/half-transparent hex-8 to hex-4 (#336699(4/8)0 -> #369(4/8)).
# + Improved: conditional at-rules formatter - fix for "one-story" at-rules have declares only (by define, e.g. @font-face, @page etc.).
# + Improved: conditional at-rules formatter - process logics changed to avoid "crooked firs" if some identical rules exist in CSS.
# + Improved: simplify media-queries (set[fine] = 2) - remove "screen" (eq. default @media all...). Use overlapped "print" rules if need.
# + Improved: extra beautifier for at-rules conditions (set[extr] = 2|3) - "or" to "," (2), max(min)-width/height to W/H <(>)= (3).
#	* Set 2|3 ONLY if u sure all media-query conditions in your CSS are correct! Otherwise some query parsing errors can be occured.
# + Improved: extra-fine numbers convertor - now used full units list (see //developer.mozilla.org/docs/Web/CSS/CSS_Values_and_Units).
# + Optional: fast/safe matches replace (set[fast]) - avoid errors with "packet" match replacement entire* CSS by mask. Default is "fast".
#	* Some cases "packet" replacement can broke correct declarations and unpredictable changes can be occured (e.g. in color condenser).
#	* If set 0: brings some slow down the conversion - all same fragments be replaced step-by-step instead of at one ops (with "fast").
# + Usability: if work "by place" (with selected text) - now after conversion caret also set and autoscrolled to the start of selection.
#
#	All versions: code optimization, simplify/shorten/speedup functions, check/fix/improve regexps, change some ops order.
#
## Requirements
#	0. NPP 6.5.1+ (this script tested in, older/newer NPP version - test yourself).
#	1. NPP Python Script plugin: install FULL version from //npppythonscript.sourceforge.net (this script runtime, required).
#	2. NPP Customize Toolbar: install from //sourceforge.net/projects/npp-customize (best usability if u often use this, optional).
#
## Installation
#	Simply copy "CSSMinMaxR24.py" to "..\plugins\PythonScript\scripts\" in your NPP folder and restart NPP.
#	It will now appear as a menu item under Plugins > Python Script > Scripts
#	For quick access u can adjust NPP Python Script plugin and add this script at top level in menu and/or on NPP toolbar.
#	For more usability u can also set the icon for this script - e.g. CSSMinMaxR24.bmp (supplied in package).
#	For best usability u can also use* the Customize Toolbar plugin to adjust menu button position/tooltip.
#	* Add the string "Plugins,Python Script,Scripts,CSSMinMaxR24,CSSMinMaxR24.bmp" in CustomizeToolbar.btn file.
#
## Usage
#	1. Select text fragment to process, or select nothing. Hint: select all (Ctrl+A) to process the whole current file "by place".
#	2. Go to NPP menu: Plugins..Python Script..Scripts..CSSMinMaxR24 and click. If u set the script toolbar button - simply click it.
#	3. Case some text selected, it will process and replace the selected fragment, otherwise the entire contents of current document.
#	4. Built-in NPP undo is available anyway - if u don't like the results. Also see options below to control autosave/undo actions.
#
## Hints
#	1. First time run (if your CSS yet not processed by this script) - run this script TWICE (to pass both compress/expand ops for your CSS).
#		* This is "control shot" to be sure any (possible) errors corrected. Simply select ALL text and apply script twice to that.
#	2. Use conversion without "extra-fine" processing only if u sure* your CSS has no errors. So, the process will passed more faster.
#		* If not - use the additional extra-fine conversion, process is some slower, but typical "kids" errors will be corrected.
#	3. Before "stream processing" all of your CSS files: test/verify conversions to be sure the script output is correct for your CSS.
#		* If u don't like some results - simply try change some settings, adjust processing flags as u wish and then test again.
#	4. For more quality warranty: u can check the source/output CSS for errors* online (e.g. at W3C CSS validator).
#		* First check raw CSS before any conversions to correct errors if any exist. After conversion recheck output CSS to be sure all ok.
#	5. If u want force recheck/recompress already compressed CSS in one step - simply add whitespace or line break* after any "{" in source.
#		* So your CSS will be force detected as "uncompressed" CSS and the compressor will be force run instead of expander.
#	6. Also u can "train on cats" with test CSS-file (included in package) to learn script works and try different options.
#
## Warning!
#	Don't switch to another tab in NPP during process runtime! Always wait for process end! Otherwise output CSS placed in tab u switch!
#	This unpredictable change is very possible on slow PC and if u work with big CSS (long time processing).
#
## Credits
#	Original cssmin.py port of YUI from //github.com/zacharyvoase/cssmin
#	Forked CSSmin.py v.1.0 from //github.com/ethanpil/npp-cssmin
#	W3C CSS validator //jigsaw.w3.org/css-validator/
#	Regex checker/validator //regex101.com/
#
## Licence
#	Original BSD changed to GPLv3 for this script and it's next childs.
#
###

## Intro ended. Are you ready? Let's go! :)
# Set options: conversion parameters, if not explained - 1 = on, 0 = off.
# Default values tested in practics, don't change em randomly if u not understand the process logic!
set = {	# NB! Five flags below are interdepends! Protected empty rules also marks as "important" comments! Be sure u not lost em on changes!
	"rmcm": 1,	# Skip, process or force remove any comments. 0: skip (stay as-is), 1: automatic with protect (default), 2: kill em all!
	"nlcm": 0,	# Comments indent with new line before/after for best readability. In compressor only, expanded CSS always has indents.
	"ciem": 0,	# Search/process comments for IE5+ on Mac. Default is 0 (useless in modern CSS). 1: only if that used/exist in your CSS.
	"imcm": "",	# Preserv important (for you) comments from remove in compressor. Default "*" (or any other symbol) = keep that.
				# DON'T USE the regexp control symbols! Set any of [^.(){}\[\]!$+/?|] break the comments processing! (and will be skipped).
	"ercm": 2,	# Save empty rules as comments. 3: automatic, 2: save all, 1: only with all commented declares, 0: remove all trash.
				# Note: if set 3 (auto) - any empty rules or commented declares will be processed by comment rules logic (selective remove).
	"bckp": 0,	# Create current CSS file backup copy before any actions (*.css.BAK).
	"pfcm": 1,	# Keep first comment: both compressor/expander, that comment (CSS "cap") always stay unchanged and not removed anyway.
	"mlvl": 1,	# Multi-level format (expander only). 1: use extra indents for embedded rules in conditional at-rules (@media/@supports/etc).
	"fine": 1,	# Experimental: CSS-beautifier, 1: expander only, increase readability, remove exseed trash, add spaces, line indents, etc.
				# 2: remove "screen" from at-rules conditions (eq. "all", can required overlapped at-rules for printed media in your CSS).
	"extr": 1,	# Experimental: extras both compressor/expander, less size, best readability, no exseeds (RGB/HEX-colors, multi-zeros, etc.)
				# 0: any extra feats NOT be applied (some increase output CSS size). 1 (and above, default): unit/color/pseudo-class fixes.
				# 2: change "or" to comma-separated media-query. 3: at-rules conditions extras (convert min/max-W/H to W/H >=/<= or to range).
		# NB! three flags below work ONLY when "extra-fine" flag NOT =0 (those experimental features is one of the "code-correctors").
	"psel": 1,	# Experimental: check/corrects missing/excess "::" in pseudo-elements (-webkit/-moz too) and excess ":" in pseudo-classes.
	"tran": 1,	# Converts "transparent" color to #0000 in compressor (and vise versa in expander). Win 6 bytes economy in compressed CSS.
	"fast": 0,	# Fast or safe search/replace. Set 0 if u has errors with "packet" replacement entire CSS by mask (e.g. in color condenser).
	"bmhk": 0,	# Use the Box Model Hack in compressor. Set to 1 ONLY if your CSS has this fix (for IE5/6, useless in modern browsers).
	"tsas": 0,	# Autosave file every time before run process. Only for "by place" ops (if works with selected text in current file).
	"undo": 0,	# Use new undo buffer (for processing "by place" only). Use 1: ONLY if u sure - avoid reset current document undo buffer!
	"wrap": 0,	# Wrap lines in compressed CSS to an approximate length (increase CSS readability). 0 = no wrap, any integer = max length.
				# Try different values to get comfortable text view - e.g. 1: very fine - one line = one rule, 0: 1-line "brick", and so on.
	"fmax": "",	# Default file extension prefix mask to autosave EXPANDED files. If empty - autoset as "_MAX", end ".css" be autoadded.
	"fmin": "",	# Default file extension prefix mask to autosave COMPRESSED files. An mass can be ".min", end ".css" will be autoadded.
	"utf8": 1	# Force convert output CSS to UTF-8 (prevent unicode in source contains non-latin symbols). 0: save with NPP settings u set.
}

if set["bckp"]: notepad.saveAsCopy((notepad.getCurrentFilename() + ".BAK"))	# At first, create source file backup... if u ON this flag... :)
tmp = ["","","","",[""]]	# Temporary strings array
from StringIO import StringIO # The pure-Python StringIO supports unicode
import re	# Regex ops module

# Remove CSS comments except protected. Remove user-defined comments with incorrect mask.
def check_comments(css):
	if re.search(r'[\^\.\(\)\{\}\[\]\*\!\$\+\/\?\|\\]', set["imcm"]): set["imcm"] = 0	# Yet as default "*" - to avoid excess checks
	deficm = "*"	# Default preserved comments sign, also used to protect empty rules from remove
	prescomm = iemac = 0; ciem = set["ciem"] and '\\*/' in css; cs = css.find("/*")
	while cs >= 0:	# Slicing to avoid an IndexError
		ce = css.find("*/", cs + 2)
		if set["rmcm"] == 1:	# Check for preserved comments like /**...*/ (by default: two stars ** followed open slash)
			chkpoint = css[cs + 2:cs + 3]
			prescomm = iemac = (1 if deficm == chkpoint or set["imcm"] == chkpoint else 	# Skip default or user-defined comments
				(1 if ciem and ce > cs and css[ce - 1] == "\\" else (2 if iemac else 0)))	# Skip IE-Mac comments and next after
		if ce < 0:
			if not prescomm: css = css[:cs]
			break
		elif ce >= (cs + 2):
			if prescomm: cs = ce + 2; tmp[1][2] = 1
			else: css = css[:cs] + css[ce + 2:]
			if iemac > 1: iemac = 0	# IE-Mac protection flag reset: when was skipped next comment after the "hacked" one
		cs = css.find("/*", cs)
	if tmp[1][2] and set["nlcm"] and set["wrap"] == 0:	# Dumbs to set comments final indents (if any exist after cleanup)
		tmp[1][1] = "___PRESCOMM___"	# Protected comments dumb, replaced to \n at last
		if set["imcm"]: css = css.replace("/*"+set["imcm"], tmp[1][1]+"/*"+set["imcm"])	# Also for user-defined comments
		css = re.sub(r"\*\/\s*", "*/"+tmp[1][1], css).replace("/**", tmp[1][1]+"/**")
	return css

# Replace things to dumbs before processing to avoid changes in literals or issues 'p :link' -> 'p:link', ') a' -> ')a'.
def no_ws_hell(css, sf, rw, rx=0, st=0):
	rw = tmp[3][0][0]+rw+tmp[3][0][1]	# Dumb template to replace/revert
	if rx:	# Replace things to dumbs
		rx = re.compile(rx); m = rx.search(css)
		while m:
			if st: sf = m.group(); tmp[3][0][st].append(sf.strip())	# Store things (revert back at last)
			css = ''.join([ css[:m.start()], m.group().replace(sf, rw), css[m.end():] ]); m = rx.search(css, m.start()+len(rw))
	else: exec("for m in tmp[3][0][st]: css = css.replace(rw, m, 1)" if st else "css = css.replace(rw, sf)")	# Revert things back
	return css

# Fast or safe matches replacement. Avoid errors in the same matches with "packet" replacement entire CSS by mask.
def fsr(css, rw, m):
	return (css.replace(m.group(), rw) if set["fast"] else css[:m.start()] + rw + css[m.end():])

# Smart remove excess whitespace characters.
def condense_spaces(css):
	css = no_ws_hell(css, " :", "3A", "(\s+:)(?=[^{]*{)")	# Dumb to prevents 'p :link' becoming 'p:link'
	css = no_ws_hell(css, ") ", "29", "\)\s+(?!(and|or))(?=[^{};]+{)")	# Dumb to prevents ') a' becoming ')a'
	css = re.sub(r"\s+([!{};:<=>)\],]|(?!(\s+-\s+))\(|\+(?!\s+[(\d-]))(?!.*(\*\/))", r"\1", css)	# Remove spaces before signs
	css = re.sub(r"([!{}:;<=>(\[,]|\)?[^\s][-+]|\)(?=\s[a-z]+))\s+", r"\1", css)	# Remove spaces after signs
	return re.sub(r"( :){2,}(\s*)", "::", no_ws_hell(no_ws_hell(css, ") ", "29"), ":", "3A"))	# Revert dumbs, fix PE-indents

# Convert rgb(a)(51,102,153(,.66))/hsl(a)(210,50%,40%(,.66)) to #336699(aa). Shorten #336699(aA) to #369(a) and #369F to #369.
def condense_colors(css):	# Process declares only: avoid possible [id]-marks or @support() conditions overlap.
	def hsl_rgb(h, s, l):	# Converts HSL-to-RGB, modified part of colorsys lib: //github.com/python/cpython/blob/3.12/Lib/colorsys.py
		def _v(m1, m2, mh):
			c23 = 2/3.0; mh = mh % 1.0
			return (m1+(m2-m1)*mh*6 if mh < 1/6.0 else (m2 if mh < 0.5 else (m1+(m2-m1)*(c23-mh)*6 if mh < c23 else m1)))
		if s == 0: return [l, l, l]
		m2 = (l*(1+s) if l <= 0.5 else l+s-(l*s)); m1 = 2*l-m2; h = h/360.0; c13 = 1/3.0
		return [_v(m1, m2, h+c13), _v(m1, m2, h), _v(m1, m2, h-c13)]

	def raw_hex(a3c, hr):	# Converts RGB-to-HEX, accuracy 0.1%, autofix syntax/logic errors
		a3c = map(lambda n: (float(n.replace('%',''))/100 if '%' in n else float(n)), filter(lambda s: re.search('\d+', s),
			re.sub(r'\s+%', '%', re.sub(r'\b(?![1-9])+0*\.+([0-9]+)', r'.\1', a3c)).replace(',', ' ').replace('/', ' ').split()))
		if len(a3c) > 3 and (a3c[3] >= 1 or round(a3c[3]*255) == 255): a3c = a3c[:3]	# Remove excess/error opacity (eq. full color)
		if hr: a3c = hsl_rgb(*a3c[:3])+([a3c[3]] if len(a3c) > 3 else [])
		return '#' + ''.join(map(lambda s: '%.2x' % int(round((float(255*s) if s < 1 else (255 if s > 255 else s)))), a3c))

	rex = re.compile(r"(rgb|hsl)a?\s*\(\s*([0-9%.,\s\/]+)\s*\)(?=[^;{}]*[;}])"); m = rex.search(css)
	while m:	# Convert /rgb(a)(20%,102,153(/.66))/hsl(a)(210,50%,40%(,.66)) to #336699(aa) (rgb/hsl-to-hex)
		css = fsr(css, raw_hex(m.group(2),'h' in m.group()), m); m = rex.search(css, m.start())
	rex = re.compile(r"(?=[^\"'=\s]\s*)#+([0-9A-F]{6,8})(?=[^;{}]*[;}])", re.I); m = rex.search(css)
	while m:	# Shorten #336699(aA/40/80) to #369(a/4/8)
		ma = re.findall(r"\w\w", m.group(1).lower()); one = [c[0] for c in ma]; two = [c[1] for c in ma]
		if one == two or (len(ma) == 4 and (ma[3] == '80' or ma[3] == '40') and one[:3] == two[:3]):
			css = fsr(css, '#' + ''.join(one), m); m = rex.search(css, m.start() + len(one) + 1)
		else: m = rex.search(css, m.end())
	rex = re.compile(r"(?=[^\"'=\s]\s*)#+([0-9A-F]{5})(?=[^\w;{}]*[;}])", re.I); m = rex.search(css)
	while m:	# Fix/revert pre-replaced 6-in-8 to 5 (if any)
		ma = m.group(1).lower()
		if ma[3] == ma[4]: css = fsr(css, '#' + ma[:4], m); m = rex.search(css, m.end() - 1)
		else: css = fsr(css, '#' + ma[0]*2 + ma[1]*2 + ma[2]*2 + ma[3] + ma[4], m); m = rex.search(css, m.end() + 3)
	rex = re.compile(r"(?=[^\"'=\s]\s*)#+([0-9A-F]{3}|[0-9A-F]{6})(F+)(?=[^\w;{}]*[;}])", re.I); m = rex.search(css)
	while m:	# Remove excess opacity bytes in hex-4/8 (if any)
		css = fsr(css, '#' + m.group(1), m); m = rex.search(css, m.end() - 2)
	if set["tran"]: css = re.sub(r""+tmp[2][4][tmp[1][0]]+"(?=[^\w;{}]*[;}])", tmp[2][4][int(not tmp[1][0])], css)
	return css

# Wrap the lines of output CSS to an approximate length.
def wrap_lines(css):
	lines = []; line_start = 0
	for i, char in enumerate(css):	# Safe to break after } symbol
		if char == '}' and (i - line_start >= set["wrap"]): lines.append(css[line_start:i + 1]); line_start = i + 1
	line_start < len(css) and lines.append(css[line_start:])
	css = '\n'.join(lines)
	return (re.sub(r"(\n*(\*\/)[\s\n]*)(?=[^{}]+{)", r"\2\n", css) if "*/" in css else css)

# Automatic fix missing/excess :: in selectors and in (webkit/moz)-declares inside the rules.
def pseudo_elements_sign(css):	# Applied ONLY if PEC flag is ON (set[psel]:1)!
	tmp[2][1] = "(after|before|marker|selection|placeholder|backdrop|highlight|cue|file-selector-button|highlight|part|slotted|target-text|first-(letter|line)|(grammar|spelling)-error|view-transition(-(image-pair|group|new|old))?)"	# //developer.mozilla.org/docs/Web/CSS/Pseudo-elements
	tmp[2][2] = "(?=.+{)"
	tmp[2][3] = "(?=(-webkit-|-moz-)[\w-])"
	css = re.sub(r"(:\s*)+((?!"+tmp[2][1]+"[^\w-])|("+tmp[2][1]+"[\w-]))"+tmp[2][2], r":\2", css)	# Pseudo-classes
	css = re.sub(r"\s*(:\s*)+(?="+tmp[2][1]+"[^\w-])"+tmp[2][2], "::", css)	# Pseudo-elements
	css = re.sub(r":+"+tmp[2][3]+tmp[2][2], "::", re.sub(r"\s?:{2,}\s*", "::", css))	# Non-standard pseudo-elements
	css = re.sub(r"([\w]:):"+tmp[2][3], (r"\1 " if tmp[1][0] else r"\1"), css)	# Correct possible errors in non-standard declares
	return re.sub(r"(\([^)]+[\w-]:):*(?=[^\s:])(?=[\w-]+[;})])", (r"\1 " if tmp[1][0] else r"\1"), css)	# Spaces in at-rules (a:b)

# Advanced post-processing: convert colors to short equivalent, condense :0 0 0 0, 0px etc. to :0, correct typical errors.
def extra_fine(css):	# Applied ONLY if extras flag is ON (set[extr]>0)!
	tmp[2][0] = ["(%|r?[dpe]*x|[dls]*v[hwbimaxn]+|[dpcmre]+m|p[tc]|r?[chiaplh]+|cq[bhimaxinw]+|deg|g?rad|m?s|in|dpi|k?[Hh][Zz]|fr|Q|turn|)", "calc\s*\([^,;{})]*([(]*-?\d+", "\s*([*/+-])", "(?=[\(\d-])", r"(background-position)(:\s*)(0)([;}])"]	# 4: Excludes: revert 0 back to 0 0
	if re.search(r"\b(and|or|(?<!:)not|calc)\s*\(", css):	# Add spaces in at-rules/calc() if omit: //developer.mozilla.org/docs/Web/CSS/calc
		css = re.sub(r"\b(and|or|(?<!:)not)\s*\(", (r" \1 (" if tmp[1][0] else r"\1 ("), css)	# In at-rules
		css = re.sub(r"\b("+tmp[2][0][1]+tmp[2][0][0]+"|\)))"+tmp[2][0][2], r"\1 \4 ", css)	# In calc(.)+(.)
		css = re.sub(r"([(]-?\d+"+tmp[2][0][0]+"|\))"+tmp[2][0][2]+tmp[2][0][3], r"\1 \3 ", css)	# In calc((.+.))
		if not tmp[1][0]: css = re.sub(r"(?="+tmp[2][0][1]+tmp[2][0][0]+")|\)?)\s+([*/])\s+", r"\3", css)	# Compressor: minify *|/ in calc()
	css = condense_colors(re.sub(r"([:\s(-])(0*\.+)(\d+)(?=[^{]*[;}])", r"\1.\3", re.sub(tmp[2][0][4], r"\1\2\3 0\4",	# 0 -> 0 0, 0.5 -> .5.
		re.sub(r"(:\s*)(0\s+){1,3}(0)([;}])", r"\1\3\4", re.sub(r"([\s:(])-*(0)+"+tmp[2][0][0]+"(?=[^.{]*[;}])", r"\1\2", css)))))	# 0px -> 0
	return (set["psel"] and pseudo_elements_sign(css) or css)	# Fix pseudo-elements/classes rules

# CSS compressor, remove useless spaces.
def cssmin(css):
	tmp[1] = [0,"","",(['"\\"}\\""', "___PSEUDOCLASSBMH___"] if set["bmhk"] else "")]	# Preset compressor flags/data
	if set["rmcm"]: css = check_comments(css)	# Check/remove CSS comments (except preserved)
	if set["bmhk"]: css = css.replace(tmp[1][3][0], tmp[1][3][1])	# Process the Box Model Hack (BMH): BMH mask and pseudo-class dumb
	css = condense_spaces(css_beauty(css, 0))	# Protect literals from any changes then search/remove excess whitespaces
	return (set["bmhk"] and css.replace(tmp[1][3][1], tmp[1][3][0]) or css)	# Revert the BMH mask

# Basic CSS formatter, parse/indents as "usual" rules only.
def base_indent(css):
	tmp[1] = [1, "___AT_COND___", "___AT_RULES_SET___", "___AT_RULES_END___", r"(['\"]data:)(.+;)(base64,)"]	# Expander flags/data
	css = re.sub(r'(\*\/\n)(?=[^{}]+})', r'\1\n\t', re.sub(r'(\*\/)', r'\1\n', css))	# Comments with new lines (also inside rules)
	css = (set["mlvl"] and re.sub(r'('+tmp[1][1]+'[^{}]+){',	# Dumbs for cobditional at-rules begin/declare/end
		r'\1'+tmp[1][2]+'\n', re.sub(tmp[2][5], tmp[1][1], css)).replace('}}', '}'+tmp[1][3]+'\n') or
		re.sub(r'(@[^{}]+{)', r'\1\n', css))	# Basic only (if multi-level flag is off): CR begin only
	css = re.sub(r'((([^({})]+)[^{}]+){)', r'\1\n\t', css)	# Rules begin
	if set["fine"] and re.search(tmp[1][4], css): css = css_beauty(css, 1)	# Best view for base64-urls
	css = re.sub(r'(;((\s*\\*\*\/)?|(\s*\\*\*\/\s*\w)))(?![^\n}])', r'\1\n\t', css)	# Next declare (also in-rule comments)
	css = re.sub(r'(?!"\'[^;]*);(?![^;\n}]*\'")', ';\n\t', css)		# Next declare
	if set["fine"]: css = css_beauty(css, 2)	# Readability: spaces before/after signs, absent ";", in-rule comments
	return re.sub(r'\t*\}\s*', '\n}\n', css)	# Rules end

# Multi-level conditional rules (CR) formatter (@media/etc. indents for inner rules).
def cond_indent(css, t1, t2, t3):
	t = "\t"; s = 1; rz = []
	for rb in css.replace(t3+';', t3).split(t1):	# Process rules inside the CR, "usual" rules with no CR skipped.
		if t3 in rb:	# Build tab-shifted "fir" from rules inside CR
			c = rb.count(t3); cr = []
			for r in rb.split(t3):
				if not r.find("\n\n}") or not ":" in r and len(r)>1: s -= 1
				cr.append((r.replace("\n", "\n"+t*s) if s>0 else r)); s -= 1
			rb = t3.join(cr); s = (s+c+1-(int(c%2==0)) if c < s else 1)	# Align "crooked fir" in next level
		elif t2 in rb: rb = rb.replace("\n", "\n"+t*s); s += 1	# Add tab-shift to child rules
		rz.append(rb)
	return "@".join(rz).replace(t2, ' {').replace('\t'+t3, '}')	# Revert back CR dumbs

# CSS beautifier: increase output CSS-text readability, remove exceeds (both compressor/expander).
def css_beauty(css, e, x=0, y="(width|height)"):
	if e == 0: return (re.search(tmp[3][1][0], css) and no_ws_hell(css, 0, "LIT", tmp[3][1][0], 2) or css)	# Dumbs to literals
	elif e == 1: return re.sub(tmp[1][4], r'\1 \2\n\t\t\3', css)	# Experimental: add tabs in base64-urls
	elif e == 2:	# Pre-improve before run conditional formatter, after all basic formats applied (avoid issues at next steps)
		css = no_ws_hell(css, ":", "3A", "(?=[^\s]):(?=[^:\d-][^{;}]+{)")	# Dumb to prevents 'p:link' becoming 'p: link'
		css = re.sub(r'([^};\s])[\n\s]*}', r'\1;\n}', css)	# Add absent ";" at last declaration end
		css = re.sub(r"(?!\s)([>+{]|!i|<(?=([=\s\d]|"+y+")))", r" \1", css)	# Add spaces before
		css = re.sub(r'((?<![\s():>]):(?=[^;{})]+[;)}])|\)(?![_,:;)}])|[,+]|>(?!=)|[<=](?=\d+|'+y+'))(?!\s)', r'\1 ', css)	# Add spaces after
		css = re.sub(r"(?=[^;,]+)((url|local)\([^;,]+,)\s*(?=[^;,]+;?)", r"\1\n\t\t", css)	# Experimental: add tabs in comma-separated urls
		return re.sub(r'(\n.+;)\s*(\\*\*\/)', r'\1 \2', no_ws_hell(css, ":", "3A"))	# Commented declares in rules by lines
	elif e == 3:	# At-rules conditions extra-finer: spaces before "(", max(min)-width(height) -> W(H) <(>)=, "or" -> ","
		c = css.split(x); s = (" " if tmp[1][0] else ""); z = ":\s*(\w+)(?=[\s)])";
		if set["fine"] > 1: c[0] = re.sub(r"(?<!not)\s*(only\s*)?screen\s*and\s*(?!\w)", "", c[0])
		c[0] = re.sub(r"([^\s(])\((?!\()", r"\1"+s+"(", (
			set["extr"] > 1 and re.sub(r"([\w)])\s*\bor\s*(?=[(\w]+)", r"\1,"+s, (
			set["extr"] > 2 and re.sub(r"([<=>])\s*"+y+r"\)\s*and\s*\(\2\s*([<>])", r"\1"+s+r"\2"+s+r"\3",
				re.sub(r"max-"+y+z, r"\1"+s+"<="+s+r"\2", re.sub(r"min-"+y+z, r"\2"+s+"<="+s+r"\1", c[0]))) or c[0])) or c[0]))
		return x.join(c)
	css = re.sub(r'[ ]{2,}', ' ', re.sub(r'([.#]){2,}(?=.+[{\w])', r'\1', css))	# Remove exceeds repeats
	if tmp[1][0]:	# Next declare after comment inside the rules. Remove exceed ";", empty lines, trailing spaces/tabs
		css = re.sub(r'(\n\s*;*)+\n', '\n', re.sub(r'\n([^;\t]+;)(\s*\})', r'\n\t\1\2', css))
	else:	# Wrap lines, indent the protected comments with new line or remove trail spaces after comments
		css = re.sub(r"[;\s]*}", "}", re.sub(r"\s+", " ", css))	# Remove trail ";", condense spaces/tabs to one
		css = (wrap_lines(css) if set["wrap"] else (css.replace(tmp[1][1], "\n") if tmp[1][1] else css.replace("*/ ", "*/")))
	return (no_ws_hell(no_ws_hell(css, 0, "LIT", 0, 2), ";", "3B") if tmp[3][0][2] else css)	# Revert literals, revert dumbs for ";"

# CSS expander/formatter: expand compressed CSS for best readability.
def cssmax(css):
	css = base_indent(css_beauty(css, 0))	# Protect literals from changes then run basic formatter (one-level "usual" rules)
	return (set["mlvl"] and tmp[1][1] in css and cond_indent(css, tmp[1][1], tmp[1][2], tmp[1][3]) or css)	# Check/run CR-formatter

# Protect/move the "cap" (first comment) and special at-rules to the BOF (@charset/@import, if exist).
def move_to_begin(css, cap=0, store=0):
	def temp_remove(css, arr): exec("for a in arr: css = css.replace(a,'',1)"); return css
	if store:	# If any special rule exist: temp store/remove from source
		tmp[0] = ["", "", ""]
		tmp[0][0] = atc = re.findall("@charset\s+['\"][\w-]+['\"]", css) or ""	# Need ONLY first @charset
		tmp[0][1] = cap = cap and cap.group() or ""	# Store first comment to avoid any changes
		tmp[0][2] = ati = re.findall("@import\s+[url(]*['\"][^'\"]+['\"]\)?[^@;{}\r\n]*(?=;?)", css) or ""	# Store all @import
		return temp_remove(temp_remove(css, atc), ati).replace(cap, "", 1)	# Temp remove "cap" and all @charset/@import
	lbs = ";"+"\n"*(tmp[1][0] or int(set["wrap"]>0))	# Revert special rules with first comment to the BOF
	return (tmp[0][0] and tmp[0][0][0]+";"+"\n"*int((len(tmp[0][1])>4 or len(lbs)>1))) + (
		tmp[0][1] and re.sub(r"\s*\n", "\n", tmp[0][1])+"\n") + (tmp[0][2] and lbs.join(tmp[0][2])+lbs) + css

# Convert source by flag max|min. Pre(post)-processing before save/replace CSS.
def outcss(css, asf, utf=0):
	asf and notepad.save()	# Autosave source before processing (if set for selected text or if target is a new file)
	tmp[2] = ["","","","", ["transparent","#0000"], r"@(?=[^{}]+{[^{}]+{)"]	# Constants array for extra-fine conversions
	tmp[3] = [["___DUMBCLASS", "___", []], [r"(?<=[:=(])\s*(?:(['\"])(?!data:.*[\n\s]*base64,)[^'\"]+\1)"]]	# Protection dumbs array
	css = re.sub(r"\/\*[\s\*]*\*\/", "", (utf and css.decode('utf8') or css))	# Force remove empty comments, anyway
	asf = set["pfcm"] and re.search(r"(^|\A)\/\*[^/]+\*\/", css, re.M) or ""	# Check for the first comment (CSS "cap")
	if asf or "@charset" in css or "@import" in css: css = move_to_begin(css, asf, 1)	# Store "cap" and @charset/@import
	css = no_ws_hell(css, ";", "3B", tmp[3][1][0])	# Dumbs to ";" in literals (content:'...'/[has("...")])
	css = re.sub("(^|[{}]);+[\n\s]*", r"\1", re.sub(r"[\s;]*;\s*", ";", css, re.S), re.M).replace(";{", "{")	# Cleanup trash/trail ";"
	if set["ercm"] < 3:	css = re.sub(set["ercm"] and r"([^{}]+{\s*)\/\*[*]*\s*([^{}]+)\*\/\s*}", (r"/** \1\2}*/" or ""),	# All commented
		re.sub(r"([^{}]+{)[\s;]*}", (set["ercm"]>1 and r"/** \1here-was:empty-rule}*/" or ""), css))	# No declares (has selector only)
	css = (re.search(r"\{[\s\n][^\*\/]", css) and cssmin(css) or cssmax(css))	# Autodetect direction: by space or line-break in any rule
	if (set["fine"]>1 or set["extr"]>1) and re.search(tmp[2][5],css): css = "@".join([css_beauty(rb,3,'{') for rb in re.split(tmp[2][5],css)])
	if set["extr"]: css = extra_fine(css)	# Advanced CSS-processing: convert values to it's short equivalent, correct typical errors
	elif re.search(r"\b(and|or|(?<!:)not|calc)\s*\(", css): css = re.sub(r"(\b(and|or|(?<!:)not)|\s[-\+])(?=\()", r"\1 ", css)	# Force spaces
	css = css_beauty(css, 4)	# Improve readability/compression (compressor/expander final ops)
	if tmp[0]: css = move_to_begin(css)	# Revert back first comment, move @charset/@import rules to the BOF
	return (utf and css.encode('utf8') or css).strip()+"\n"	# Remove trails, add line at the EOF

## Functions defined. Now execute: selected text processed "by place", otherwise - entire document.
css = editor.getSelText()	# Get selected
if css:	# If selection in current document: convert fragment "by place"
	set["undo"] and editor.beginUndoAction()	# Packet undo: add ops to current document undo buffer
	editor.targetFromSelection(); editor.replaceTarget(outcss(css, set["tsas"]))	# Autosave before process selected text (if set)
	set["undo"] and editor.endUndoAction()
	editor.setSel(editor.getTargetEnd(), editor.getTargetStart())	# Usability: re-select output CSS fragment after conversion
else:	# If nothing selected: convert whole current CSS then save as new file, output file prefix autoset by conversion direction (min/max)
	tmp[4] = [notepad.getCurrentFilename(), outcss(editor.getCharacterPointer(), 1, set["utf8"])] 	# Autosave before create new file
	notepad.new(); set["utf8"] and editor.setCodePage(65001); editor.replaceTarget(tmp[4][1])	# Usability: stay at top of created file
	notepad.saveAs(re.sub(r"\.css$", (tmp[1][0] and (set["fmax"] and set["fmax"] or "_MAX") or set["fmin"] and set["fmin"] or "_")+".css", tmp[4][0]))

tmp = ""	# Reset temp-strings

# That's all, folks! :)
